@@ -27,13 +27,15 @@ rlpython {}
 
 
 class Repl:
-    def __init__(self, banner=DEFAULT_BANNER, prompt=DEFAULT_PROMPT,
+    def __init__(self, banner=DEFAULT_BANNER, warnings=[],
+                 prompt=DEFAULT_PROMPT,
                  prompt_ps2=DEFAULT_PROMPT_PS2,
                  history_file=DEFAULT_HISTORY_FILE,
                  history_size=DEFAULT_HISTORY_SIZE,
-                 globals=None, locals=None):
+                 globals={}, locals={}):
 
         self.banner = banner
+        self.warnings = warnings
         self.ps1 = prompt
         self.ps2 = prompt_ps2
         self.history_file = os.path.expanduser(history_file)
@@ -42,6 +44,7 @@ class Repl:
         self.exit_code = 0
 
         # setup readline
+        readline.set_auto_history(False)
         readline.parse_and_bind('tab: complete')
         readline.set_completer(self.complete)
 
@@ -85,9 +88,6 @@ class Repl:
         # rotate history
         self.history = self.history[self.history_size*-1:]
 
-        # reset readline history
-        readline.clear_history()
-
         for item in self.history:
             readline.add_history(item)
 
@@ -101,14 +101,24 @@ class Repl:
             pass
 
     # error management ########################################################
-    def write_exception(self, exception):
+    def write_exception(self, exception, prefix=''):
         lines = format_exception(
             etype=type(exception),
             value=exception,
             tb=exception.__traceback__,
         )
 
-        self.write(color(''.join(lines), color='red'))
+        text = ''.join(lines)
+
+        if prefix:
+            lines = text.splitlines()
+
+            for index, line in enumerate(lines):
+                lines[index] = prefix + line
+
+            text = '\n'.join(lines) + '\n'
+
+        self.write(color(text, color='red'))
 
     # line buffer #############################################################
     def clear_line_buffer(self):
@@ -146,7 +156,7 @@ class Repl:
 
         return ps2
 
-    def embed(self):
+    def interact(self):
         while True:
             try:
                 self.line_buffer += input(self._gen_prompt()) + '\n'
@@ -163,9 +173,16 @@ class Repl:
                 continue
 
             except EOFError:
-                self.handle_ctrl_d()
+                if self.handle_ctrl_d():
+                    return
 
-                return
+                continue
+
+            # emulated CTRL-C
+            if self.line_buffer.strip().endswith('!'):
+                self.handle_ctrl_c()
+
+                continue
 
             # multi line
             if not self._validate_line_buffer():
@@ -182,20 +199,46 @@ class Repl:
     def setup(self):
         self.write(self.banner)
 
+        for warning in self.warnings:
+            self.write(color('WARNING: {}\n'.format(warning), color='yellow'))
+
     def complete(self, text, state):
         return self.python_runtime.complete(text, state)
 
     def handle_ctrl_d(self):
-        self.write('^D\n')
+        self.clear_line_buffer()
+        self.write('\n')
+
+        while True:
+            try:
+                answer = input('Do you really want to exit ([y]/n)? ')
+                answer = answer.strip().lower()
+
+                if answer in ('y', ''):
+                    return True
+
+                if answer == 'n':
+                    return False
+
+            except(EOFError, KeyboardInterrupt):
+                self.write('\n')
+
+                return False
 
     def handle_ctrl_c(self):
-        self.write('^C\n')
-
         self.clear_line_buffer()
+        self.write('^C\n')
+        self.exit_code = 1
 
     def run(self):
-        if self.line_buffer.startswith('%'):
-            self.exit_code = self.shell_runtime.run(self.line_buffer)
+        try:
+            if self.line_buffer.startswith('%'):
+                self.exit_code = self.shell_runtime.run(self.line_buffer)
 
-        else:
-            self.exit_code = self.python_runtime.run(self.line_buffer)
+            else:
+                self.exit_code = self.python_runtime.run(self.line_buffer)
+
+        except Exception as exception:
+            self.exit_code = 1
+
+            self.write_exception(exception, prefix='rlpython: ')
